from dagster import DagsterType
from qdk.utils.typing import is_dataframe, is_series, is_model, is_mlflow_run


DataFrameType = DagsterType(
    name="DataFrameType",
    type_check_fn=lambda _, value: is_dataframe(value),
    description="Can represent either a pandas or dask dataframe.",
)
SeriesType = DagsterType(
    name="SeriesType",
    type_check_fn=lambda _, value: is_series(value),
    description="Can represent either a pandas or dask series.",
)
ModelType = DagsterType(
    name="ModelType",
    type_check_fn=lambda _, value: is_model(value),
    description="Generic typing for machine learning models.",
)
MLFlowRunType = DagsterType(
    name="MLFlowRunType",
    type_check_fn=lambda _, value: is_mlflow_run(value),
    description="A type that represents a qdk.MLFlowRun class.",
)



from dagster.config.field import Field
import pandas as pd
import dask.dataframe as dd
from dagster import InputDefinition, Permissive
from qdk.dagster_types import DataFrameType, SeriesType
from qdk.training.base import TrainingComponent
from typing import Any, Dict, Union
from sklearn.base import BaseEstimator

# from dask_ml.model_selection import RandomizedSearchCV
from sklearn.model_selection import RandomizedSearchCV
import mlflow


class GridSearchTrainingComponent(TrainingComponent):
    config_schema = {
        "params": Field(
            Permissive({}),
            description="The grid search parameters to provide to the model",
        ),
        "scoring": Field(
            str,
            description="The scoring function to use for the grid search",
        ),
        "n_iter": Field(
            int,
            default_value=10,
            description="The number of iterations to run the grid search",
        ),
    }

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series],
        model: BaseEstimator,
        params: Dict[str, Any],
        scoring: str,
        n_iter: int = 10,
    ) -> BaseEstimator:
        # Create the search object
        search = RandomizedSearchCV(
            estimator=model,
            param_distributions=params,
            n_iter=n_iter,
            random_state=0,
            scoring=scoring,
        )

        mlflow.sklearn.autolog()

        # Fit the search on the data
        with mlflow.start_run() as run:
            search.fit(X, y)

        return search



from typing import List, Union

from dagster import Enum, EnumValue, Field
from dask.dataframe import DataFrame as DaskDataFrame
from dask.dataframe import Series as DaskSeries
from dask.dataframe.core import DataFrame
from pandas import DataFrame, Series
from qdk.transform.base import TransformComponent
from stop_words import get_stop_words


class TokenizeComponent(TransformComponent):
    config_schema = {
        "text_columns": Field(
            list,
            default_value=["text"],
            description="Columns to tokenize",
        ),
        "language": Field(
            Enum(
                "Language",
                [
                    EnumValue("en"),
                    EnumValue("nl"),
                ],
            ),
            default_value="nl",
            description="The language of the input text. Used to remove stopwords",
        ),
        "lower": Field(
            bool,
            default_value=True,
            description="Whether to transform the text to lowercase",
        ),
        "remove_numbers": Field(
            bool,
            default_value=True,
            description="Whether to remove numbers from the text",
        ),
        "min_length": Field(
            int,
            default_value=3,
            description="Minimum length of a token",
        ),
        "remove_stop_words": Field(
            bool,
            default_value=True,
            description="Whether to remove stopwords from the text",
        ),
        "join_text": Field(
            bool,
            default_value=False,
            description="Whether join the tokens into one string",
        ),
    }

    @staticmethod
    def _clean_tokens(
        tokens: List[str],
        min_length: int,
        stop_words: List[str],
    ) -> List[str]:
        """Cleans the tokens by removing stop words and short tokens.

        Args:
            tokens (List[str]): The tokens to clean.
            min_length (int): The minimum length of a token to keep.
            stop_words (List[str]): The list of stop words to remove.

        Returns:
            List[str]: The cleaned tokens.
        """
        stop_words_set = set(stop_words)
        return [
            token
            for token in tokens
            if token
            and len(token) >= min_length
            and token.lower() not in stop_words_set
        ]

    @classmethod
    def _tokenize(
        cls,
        series: Union[Series, DaskSeries],
        stop_words: List[str],
        lower: bool,
        remove_numbers: bool,
        min_length: int,
        join_text: bool,
    ) -> Union[Series, DaskSeries]:
        """Function to tokenize a text column.

        Args:
            series (Union[Series, DaskSeries]): The series to tokenize.
            stop_words (List[str]): The list of stop words to remove.
            lower (bool): Whether to transform the text to lowercase.
            remove_numbers (bool): Whether to remove numbers.
            min_length (int): The minimum length of a token.
            join_text (bool): Whether to rejoin the tokens into a string.

        Returns:
            Union[Series, DaskSeries]: The tokenized series.
        """
        # Replace all NaN values with an empty string
        series = series.fillna("")

        # Conditionally lowercase the text
        if lower:
            series = series.str.lower()

        # Remove numbers
        if remove_numbers:
            series = series.str.replace(r"[0-9]", "", regex=True)

        # Tokenize the text
        series = series.str.split(r"\W+")

        # Remove empty tokens
        # Minimum length of tokens
        # Remove stop words
        series = series.map(
            lambda tokens: cls._clean_tokens(tokens, min_length, stop_words)
        )

        # Optionally rejoin the text into a string
        if join_text:
            series = series.str.join(" ")

        return series

    @classmethod
    def transform(
        cls,
        df: Union[DataFrame, DaskDataFrame],
        text_columns: List[str] = ["text"],
        language: str = "nl",
        lower: bool = True,
        remove_numbers: bool = True,
        min_length: int = 3,
        remove_stop_words: bool = True,
        join_text: bool = False,
    ) -> Union[DataFrame, DaskDataFrame]:
        """
        Tokenize certain columns in a dataframe.

        Args:
            df (Union[DataFrame, DaskDataFrame]): The dataframe that contains the text columns.
            text_columns (List[str], optional): The names of the textual columns. Defaults to ["text"].
            language (str, optional): The language of the text, used for stopword removal. Defaults to "nl".
            lower (bool, optional): Whether to lower the text. Defaults to True.
            remove_numbers (bool, optional): Whether to remove numbers. Defaults to True.
            min_length (int, optional): The minimum length of the tokens. Defaults to 3.
            remove_stop_words (bool, optional): Whether to remove stopwords. Defaults to True.
            join_text (bool): Whether to rejoin the tokens into a string. Default to False

        Returns:
            Union[DataFrame, DaskDataFrame]: Return the dataframe with the tokenized columns. The columns are prefixed with "_tokens_".
        """
        # Get stop words for the language specified
        stop_words = get_stop_words(language=language) if remove_stop_words else []

        # Tokenize the text columns specified
        for text_column in text_columns:
            # Get the text column and tokenize it
            df[f"_tokens_{text_column}"] = cls._tokenize(
                series=df[text_column],
                stop_words=stop_words,
                lower=lower,
                remove_numbers=remove_numbers,
                min_length=min_length,
                join_text=join_text,
            )

        return df



from typing import Union

import dask.dataframe as dd
import pandas as pd
from dagster import InputDefinition, OutputDefinition
from qdk.base import BaseComponent
from qdk.dagster_types import DataFrameType, ModelType, SeriesType

from sklearn.base import BaseEstimator


class TrainingComponent(BaseComponent):
    compute_function = "train"
    tags = {"kind": "training"}
    input_defs = [
        InputDefinition("X", DataFrameType),
        InputDefinition("y", SeriesType),
        InputDefinition("model", ModelType),
    ]
    output_defs = [OutputDefinition(ModelType, "model")]

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series],
        model: BaseEstimator,
    ) -> BaseEstimator:
        raise NotImplementedError(
            'Make sure you added a "train" function to the component'
        )



from .transform.base import TransformComponent
from .transform.tokenize import TokenizeComponent
from .transform.yake import YakeComponent
from .transform.train_test import TrainTestComponent
from .transform.array import ArrayTransformer

from .inference.base import InferenceComponent
from .inference.mlflow import MLFlowInferenceComponent

from .training.base import TrainingComponent
from .training.mlflow import MLFlowTrainingComponent
from .training.sklearn import SklearnComponent
from .training.grid_search import GridSearchTrainingComponent

from .dagster_types import DataFrameType, SeriesType, ModelType, MLFlowRunType

from .loader.dataframe import DataFrameLoader

from .resources.io_manager import io_manager



from dagster.core.storage.io_manager import IOManager, io_manager as dagster_io_manager
from qdk.materialization import Materializer


class QDKIOManager(IOManager):
    def __init__(self):
        self.values = {}

    def handle_output(self, context, obj):
        keys = tuple(context.get_output_identifier())
        self.values[keys] = obj

        # Create a materializer for the object
        asset_materialization = Materializer(
            asset_key=[
                context.pipeline_name,
                context.solid_def.name,
                keys[-1],
            ],
            object=obj,
        ).materialize()

        # If the asset was materialized, yield the asset
        if asset_materialization:
            yield asset_materialization

        print(keys, type(obj))

    def load_input(self, context):
        keys = tuple(context.upstream_output.get_output_identifier())
        obj = self.values[keys]
        return obj


@dagster_io_manager
def io_manager(_):
    return QDKIOManager()
